// Copyright 2008 - SIsécurité
//
// Parts are adapted from Bob the Butcher.
//
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <spu_intrinsics.h>
#include "nt_spu.h"
#include "charset.h"
#include "gen_plaintext.h"

// Hold password
static unsigned char password[MAX_PWD_LEN + 1] __attribute__ ((aligned (16)));
// End position
static int end __attribute__((aligned (16)));


static void PwdInit(unsigned char *password,
		    const int start_pos) {
  int i;

  if (start_pos == 0)
    return;

  for (i = 0; i <= start_pos; ++i)
    password[i] = incrementer[0];
}


// Code adapted from Bob the Butcher
static unsigned short PwdUpdate(unsigned char *password) {
  unsigned short i = 0;

  while ((password[i] = incrementer[(unsigned int) password[i]]) == 0) {
    password[i++] = incrementer[0];

    if(password[i] == 0) {
      password[i] = incrementer[0];
      return i;
    }
  }

  return i;
}


// Will generate passwords between [start_pos, end_pos]
// e.g. w[or]d  <=> [1, 2]
void GenInit(const int start_pos,
	     const int end_pos,
	     const unsigned int charset_type,
	     const unsigned int max_len) {
  assert(max_len <= MAX_PWD_LEN);

  // Init idx2chr[] and incrementer() for charset_type
  IncrementerInit(charset_type);

  // Set all bits to 0
  memset(password, 0, MAX_PWD_LEN + 1);

  // Set end position
  if (end_pos == -1)
    end = max_len - 1;
  else
    end = end_pos;

  // Eventually, sanity check accepted range lengths
  assert(start_pos <= end);
  assert(end >= 0);
  assert((unsigned int) end < max_len);

  PwdInit(password, start_pos);
}


int DummyGeneratePlainText(unsigned char *buffer,
			   const unsigned int num,
			   const unsigned int len,
			   void *n __attribute__ ((unused))) {
  unsigned int i;
  int pos;

  assert(!(num % NUM_INTER_PT));

  for (i = 0; i < num; ++i) {
    memcpy(buffer + i * len, password, len);
    pos = PwdUpdate(password);
    if (pos > end)
      return 0;  // Stop condition met
  }

  // Even if the previous step was unable to generate num
  // passwords, it doesn't matter because at worst the remaining
  // slots will contain old data.
  return num;
}


int DummyGeneratePlainText8(unsigned char *buffer,
			    const unsigned int num,
			    const unsigned int len,
			    void *n __attribute__ ((unused))) {
  vec_ushort8 rvec = {0};

  assert(num == 8);

  // Even if the stop condition is met before having
  // finished to generate 8 passwords, it doesn't
  // really matter, at worst those password will be
  // truncated.
  memcpy(buffer, password, len);
  rvec = spu_insert(PwdUpdate(password), rvec, 0);
  memcpy(buffer + len, password, len);
  rvec = spu_insert(PwdUpdate(password), rvec, 1);
  memcpy(buffer + 2 * len, password, len);
  rvec = spu_insert(PwdUpdate(password), rvec, 2);
  memcpy(buffer + 3 * len, password, len);
  rvec = spu_insert(PwdUpdate(password), rvec, 3);
  memcpy(buffer + 4 * len, password, len);
  rvec = spu_insert(PwdUpdate(password), rvec, 4);
  memcpy(buffer + 5 * len, password, len);
  rvec = spu_insert(PwdUpdate(password), rvec, 5);
  memcpy(buffer + 6 * len, password, len);
  rvec = spu_insert(PwdUpdate(password), rvec, 6);
  memcpy(buffer + 7 * len, password, len);
  rvec = spu_insert(PwdUpdate(password), rvec, 7);

  if (spu_extract(spu_gather(spu_cmpgt(rvec, end)), 0))
    return 0; // Stop condition met
  else
    return num;
}


void PwdPrint() {
  printf("%s\n", password);
}


void PwdCopy(unsigned char *to, unsigned int len) {
  strncpy((char *) to, (char *) password, len);
}
