// Copyright 2008 - SIscurit
//
// NT MD4 accelerated encrypt code.
//
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <spu_intrinsics.h>
#include <libmisc.h>
#include "jtr_atoi.h"
#include "spu_common.h"
#include "nt_spu.h"
#include "tests.h"
#include "gen_plaintext.h"
#include "wordlist.h"
#include "parse_opt.h"


// 128 bytes aligned data seems a little better performing than with
// 16 bytes alignment.

// Store matchs
static unsigned char *results[MAX_CIPHERTEXTS] __attribute__ ((aligned (128)));

// Holds entry (hex string) ciphertexts
static unsigned char *s_ciphertexts[MAX_CIPHERTEXTS + 1] __attribute__ ((aligned (128))) = {NULL};

// Holds current ciphertexts
static unsigned int b_ciphertexts[OUT_BINARY_LENGTH * MAX_CIPHERTEXTS] __attribute__ ((aligned (128)));

// Holds currents strings plaintexts
static unsigned char s_ptext_buf[NUM_INTER_PT * IN_UTF8_LENGTH] __attribute__ ((aligned (128)));

// Intermediary plaintexts holder
static vec_uint4 i_ptext_buf[(NUM_INTER_PT * IN_UTF8_LENGTH) >> 2] __attribute__ ((aligned (128)));

// Holds current plaintexts
static vec_uint4 b_ptext_buf[(NUM_INTER_PT >> 2) * IN_BINARY_LENGTH] __attribute__ ((aligned (128)));

/* Store results */
static vec_uint4 output[NUM_INTER_PT] __attribute__ ((aligned (128)));


/* Init consts */
static const vec_uint4 init_a = {0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff};
static const vec_uint4 init_b = {0xefcdab89, 0xefcdab89, 0xefcdab89, 0xefcdab89};
static const vec_uint4 init_c = {0x98badcfe, 0x98badcfe, 0x98badcfe, 0x98badcfe};
static const vec_uint4 init_d = {0x10325476, 0x10325476, 0x10325476, 0x10325476};

/* Consts used in round 2 and 3 */
static const vec_uint4 stage_2 = {0x5a827999, 0x5a827999, 0x5a827999, 0x5a827999};
static const vec_uint4 stage_3 = {0x6ed9eba1, 0x6ed9eba1, 0x6ed9eba1, 0x6ed9eba1};


#define STEP1(A1, B1, C1, D1, A2, B2, C2, D2, X, S) {			\
    A1 = spu_add(A1, b_ptext_buf[(NUM_INTER_PT / 4) * X]);		\
    A2 = spu_add(A2, b_ptext_buf[(NUM_INTER_PT / 4) * X + 1]);		\
    t1 = spu_xor(C1, D1);						\
    t3 = spu_xor(C2, D2);						\
    t1 = spu_and(B1, t1);						\
    t3 = spu_and(B2, t3);						\
    t1 = spu_xor(D1, t1);						\
    t3 = spu_xor(D2, t3);						\
    A1 = spu_add(A1, t1);						\
    A2 = spu_add(A2, t3);						\
    t1 = spu_sl(A1, S);							\
    t3 = spu_sl(A2, S);							\
    A1 = spu_rlmask(A1, S - 32);					\
    A2 = spu_rlmask(A2, S - 32);					\
    A1 = spu_or(t1, A1);						\
    A2 = spu_or(t3, A2);						\
  }


#define STEP2(A1, B1, C1, D1, A2, B2, C2, D2, X, S) {			\
    A1 = spu_add(A1, b_ptext_buf[(NUM_INTER_PT / 4) * X]);		\
    A2 = spu_add(A2, b_ptext_buf[(NUM_INTER_PT / 4) * X + 1]);		\
    t1 = spu_or(C1, D1);						\
    t3 = spu_or(C2, D2);						\
    t2 = spu_and(C1, D1);						\
    t4 = spu_and(C2, D2);						\
    t1 = spu_and(B1, t1);						\
    t3 = spu_and(B2, t3);						\
    A1 = spu_add(A1, stage_2);						\
    A2 = spu_add(A2, stage_2);						\
    t1 = spu_or(t1, t2);						\
    t3 = spu_or(t3, t4);						\
    A1 = spu_add(A1, t1);						\
    A2 = spu_add(A2, t3);						\
    t1 = spu_sl(A1, S);							\
    t3 = spu_sl(A2, S);							\
    A1 = spu_rlmask(A1, S - 32);					\
    A2 = spu_rlmask(A2, S - 32);					\
    A1 = spu_or(t1, A1);						\
    A2 = spu_or(t3, A2);						\
  }


#define STEP3(A1, B1, C1, D1, A2, B2, C2, D2, X, S) {			\
    A1 = spu_add(A1, b_ptext_buf[(NUM_INTER_PT / 4) * X]);		\
    A2 = spu_add(A2, b_ptext_buf[(NUM_INTER_PT / 4) * X + 1]);		\
    t1 = spu_xor(C1, D1);						\
    t3 = spu_xor(C2, D2);						\
    A1 = spu_add(A1, stage_3);						\
    A2 = spu_add(A2, stage_3);						\
    t1 = spu_xor(B1, t1);						\
    t3 = spu_xor(B2, t3);						\
    A1 = spu_add(A1, t1);						\
    A2 = spu_add(A2, t3);						\
    t1 = spu_sl(A1, S);							\
    t3 = spu_sl(A2, S);							\
    A1 = spu_rlmask(A1, S - 32);					\
    A2 = spu_rlmask(A2, S - 32);					\
    A1 = spu_or(t1, A1);						\
    A2 = spu_or(t3, A2);						\
  }


void NTMD4Crypt() {
  vec_uint4 a1, a2, b1, b2, c1, c2, d1 ,d2;
  vec_uint4 t1, t2, t3, t4;
  a1 = a2 = init_a;
  b1 = b2 = init_b;
  c1 = c2 = init_c;
  d1 = d2 = init_d;
  t1 = t2 = t3 = t4 = (vec_uint4) {0, 0, 0, 0};

  /* round 1 */
  a1 = spu_add(a1, b_ptext_buf[0]);
  a2 = spu_add(a2, b_ptext_buf[1]);
  a1 = spu_sl(a1, 3);
  a2 = spu_sl(a2, 3);
  STEP1(d1, a1, b1, c1, d2, a2, b2, c2, 1 , 7 );
  STEP1(c1, d1, a1, b1, c2, d2, a2, b2, 2 , 11);
  STEP1(b1, c1, d1, a1, b2, c2, d2, a2, 3 , 19);
  STEP1(a1, b1, c1, d1, a2, b2, c2, d2, 4 , 3 );
  STEP1(d1, a1, b1, c1, d2, a2, b2, c2, 5 , 7 );
  STEP1(c1, d1, a1, b1, c2, d2, a2, b2, 6 , 11);
  STEP1(b1, c1, d1, a1, b2, c2, d2, a2, 7 , 19);
  STEP1(a1, b1, c1, d1, a2, b2, c2, d2, 8 , 3 );
  STEP1(d1, a1, b1, c1, d2, a2, b2, c2, 9 , 7 );
  STEP1(c1, d1, a1, b1, c2, d2, a2, b2, 10, 11);
  STEP1(b1, c1, d1, a1, b2, c2, d2, a2, 11, 19);
  STEP1(a1, b1, c1, d1, a2, b2, c2, d2, 12, 3 );
  STEP1(d1, a1, b1, c1, d2, a2, b2, c2, 13, 7 );
  STEP1(c1, d1, a1, b1, c2, d2, a2, b2, 14, 11);
  STEP1(b1, c1, d1, a1, b2, c2, d2, a2, 15, 19);

  /* round 2 */
  STEP2(a1, b1, c1, d1, a2, b2, c2, d2, 0 , 3 );
  STEP2(d1, a1, b1, c1, d2, a2, b2, c2, 4 , 5 );
  STEP2(c1, d1, a1, b1, c2, d2, a2, b2, 8 , 9 );
  STEP2(b1, c1, d1, a1, b2, c2, d2, a2, 12, 13);
  STEP2(a1, b1, c1, d1, a2, b2, c2, d2, 1 , 3 );
  STEP2(d1, a1, b1, c1, d2, a2, b2, c2, 5 , 5 );
  STEP2(c1, d1, a1, b1, c2, d2, a2, b2, 9 , 9 );
  STEP2(b1, c1, d1, a1, b2, c2, d2, a2, 13, 13);
  STEP2(a1, b1, c1, d1, a2, b2, c2, d2, 2 , 3 );
  STEP2(d1, a1, b1, c1, d2, a2, b2, c2, 6 , 5 );
  STEP2(c1, d1, a1, b1, c2, d2, a2, b2, 10, 9 );
  STEP2(b1, c1, d1, a1, b2, c2, d2, a2, 14, 13);
  STEP2(a1, b1, c1, d1, a2, b2, c2, d2, 3 , 3 );
  STEP2(d1, a1, b1, c1, d2, a2, b2, c2, 7 , 5 );
  STEP2(c1, d1, a1, b1, c2, d2, a2, b2, 11, 9 );
  STEP2(b1, c1, d1, a1, b2, c2, d2, a2, 15, 13);

  /* round 3 */
  STEP3(a1, b1, c1, d1, a2, b2, c2, d2, 0 , 3 );
  STEP3(d1, a1, b1, c1, d2, a2, b2, c2, 8 , 9 );
  STEP3(c1, d1, a1, b1, c2, d2, a2, b2, 4 , 11);
  STEP3(b1, c1, d1, a1, b2, c2, d2, a2, 12, 15);
  STEP3(a1, b1, c1, d1, a2, b2, c2, d2, 2 , 3 );
  STEP3(d1, a1, b1, c1, d2, a2, b2, c2, 10, 9 );
  STEP3(c1, d1, a1, b1, c2, d2, a2, b2, 6 , 11);
  STEP3(b1, c1, d1, a1, b2, c2, d2, a2, 14, 15);
  STEP3(a1, b1, c1, d1, a2, b2, c2, d2, 1 , 3 );
  STEP3(d1, a1, b1, c1, d2, a2, b2, c2, 9 , 9 );
  STEP3(c1, d1, a1, b1, c2, d2, a2, b2, 5 , 11);

  /*
  STEP3(b1, c1, d1, a1, b2, c2, d2, a2, 13, 15);
  STEP3(a1, b1, c1, d1, a2, b2, c2, d2, 3 , 3 );
  STEP3(d1, a1, b1, c1, d2, a2, b2, c2, 11, 9 );
  STEP3(c1, d1, a1, b1, c2, d2, a2, b2, 7 , 11);
  STEP3(b1, c1, d1, a1, b2, c2, d2, a2, 15, 15);
  */

  // To speed things up few steps are discarded.
  b1 = spu_add(b_ptext_buf[26], b1);
  b2 = spu_add(b_ptext_buf[27], b2);
  t1 = spu_xor(d1, a1);
  t2 = spu_xor(d2, a2);
  t1 = spu_xor(c1, t1);
  t2 = spu_xor(c2, t2);
  b1 = spu_add(t1, b1);
  b2 = spu_add(t2, b2);

  /* result */
  /* output[((NUM_INTER_PT / 4) * X) + 0] */
  output[0] = a1;
  output[1] = a2;
  output[2] = b1;
  output[3] = b2;
  output[4] = c1;
  output[5] = c2;
  output[6] = d1;
  output[7] = d2;
}


static void DisplayOutput() {
  unsigned int i;

  printf("\nOutput:\n");
  for (i = 0; i < NUM_INTER_PT; ++i)
    DisplayUIntVec(output[i]);
}


static unsigned int CmpEncryptedDeep(unsigned char *results[],
				     unsigned char *plaintexts,
				     const unsigned int *in_ct,
				     const unsigned int num_ct) {

  // First, crypt work must be completed, crypt func shortened
  // last steps.
  unsigned int pos1, pos2, pos3;
  vec_uint4 t1, t2, t3, t4;
  unsigned short true_match;
  unsigned int match = 0;

  t2 = output[2];
  t4 = output[3];
  output[2] = spu_add(output[2], stage_3);
  output[3] = spu_add(output[3], stage_3);
  t1 = spu_sl(output[2], 15);
  t3 = spu_sl(output[3], 15);
  output[2] = spu_rlmask(output[2], -17);
  output[3] = spu_rlmask(output[3], -17);
  output[2] = spu_or(t1, output[2]);
  output[3] = spu_or(t3, output[3]);

  STEP3(output[0], output[2], output[4], output[6],
	output[1], output[3], output[5], output[7], 3 , 3 );
  STEP3(output[6], output[0], output[2], output[4],
	output[7], output[1], output[3], output[5], 11, 9 );
  STEP3(output[4], output[6], output[0], output[2],
	output[5], output[7], output[1], output[3], 7 , 11);
  output[2] = t2;
  output[3] = t4;

  // Then compare every 4 words outputs with every ciphertexts.
  // This is a highly unlikely path thus it doesn't need to be
  // optimized.
  for (pos1 = 0; pos1 < NUM_INTER_PT; ++pos1)
    for (pos2 = 0; pos2 < num_ct; ++pos2) {
      true_match = 1;

      for (pos3 = 0; pos3 < 4; ++pos3)
	if (spu_extract(output[(pos1 >> 2) + (pos3 << 1)],
			pos1 % 4) != in_ct[(pos2 << 2) + pos3]) {
	  true_match = 0;
	  break;
	}

      if (true_match) {
	if (results[pos2] == NULL) {
	  match = 1;
	  assert(pos2 < MAX_CIPHERTEXTS);
	  assert(pos1 < NUM_INTER_PT);

	  results[pos2] = (unsigned char *) malloc_align(IN_CT_LENGTH * sizeof(char), 4);
	  strncpy((char *) results[pos2],
		  (char *) &plaintexts[pos1 * IN_UTF8_LENGTH],
		  IN_UTF8_LENGTH);
	  results[pos2][IN_UTF8_LENGTH] = '\0';
	}
      }
    }

  return match;
}


// Requirements:
// * NUM_INTER_PT == 8
// * (num_ct % 2) == 0
//
// This function only compare second word of every outputs to every
// ciphertext. Should be fast because heavily used.
// If a potential match is found appropriate actions will be taken.
static unsigned short CmpEncryptedShallow(const unsigned int *in_ct,
					  const unsigned int num_ct) {
  unsigned match1, match2, match3, match4, i;
  vec_uint4 v1, v2, v3, v4;
  unsigned short result = 0;

  for (i = 0; i < num_ct; i += 2) {
    v1 = spu_cmpeq(output[2], in_ct[(i << 2) + 1]);
    v2 = spu_cmpeq(output[3], in_ct[(i << 2) + 1]);
    v3 = spu_cmpeq(output[2], in_ct[((i + 1) << 2) + 1]);
    v4 = spu_cmpeq(output[3], in_ct[((i + 1) << 2) + 1]);
    v1 = spu_gather(v1);
    v2 = spu_gather(v2);
    v3 = spu_gather(v3);
    v4 = spu_gather(v4);
    match1 = spu_extract(v1, 0);
    match2 = spu_extract(v2, 0);
    match3 = spu_extract(v3, 0);
    match4 = spu_extract(v4, 0);

    // Fixme: used for debugging
    //result |= (match1 > 32) || (match2 > 32) || (match3 > 32) || (match4 > 32);
    result |= (match1 > 0) || (match2 > 0) || (match3 > 0) || (match4 > 0);
  }

  return result;
}


static void PreparePlainTexts(vec_uint4 *dst,
			      const unsigned char *in_ptext) {
  unsigned int i, j;
  vec_uchar16 *vtmp;
  vec_uint4 *p_dst;

  memset(dst, 0, sizeof(unsigned int) * NUM_INTER_PT * IN_UTF8_LENGTH);

  for (i = 0; i < NUM_INTER_PT; ++i) {
    vtmp = (vec_uchar16 *) &in_ptext[i * IN_UTF8_LENGTH];
    p_dst = dst + (i >> 2);

    for (j = 0; spu_extract(vtmp[j >> 4], j % 16) && j < IN_UTF8_LENGTH;
	 ++j, p_dst += (NUM_INTER_PT >> 2))
      *p_dst = spu_insert(spu_extract(vtmp[j >> 4], j % 16), *p_dst, i % 4);

    *p_dst = spu_insert(0x80, *p_dst, i % 4);
    p_dst = &dst[((IN_UTF8_LENGTH - 4) * NUM_INTER_PT + i) >> 2];
    *p_dst = spu_insert(j << 4, *p_dst, i % 4);
    p_dst += NUM_INTER_PT;
    *p_dst = spu_insert(0, *p_dst, i % 4);
  }
}


static void ConvertPlainTexts(vec_uint4 *in_ptext) {
  unsigned int i, index1, index2;
  vec_uint4 *t1, *t2, *t3, *t4;

  for (i = 0; i < IN_BINARY_LENGTH; i += 2) {
    t1 = in_ptext++;
    t2 = in_ptext++;
    index1 = (NUM_INTER_PT * i) >> 2;
    b_ptext_buf[index1] = spu_rl(*in_ptext++, 16);
    b_ptext_buf[index1 + 1] = spu_rl(*in_ptext++, 16);
    t3 = in_ptext++;
    t4 = in_ptext++;
    index2 = (NUM_INTER_PT >> 2) * (i + 1);
    b_ptext_buf[index2] = spu_rl(*in_ptext++, 16);
    b_ptext_buf[index2 + 1] = spu_rl(*in_ptext++, 16);
    b_ptext_buf[index1] = spu_or(*t1, b_ptext_buf[index1]);
    b_ptext_buf[index1 + 1] = spu_or(*t2, b_ptext_buf[index1 + 1]);
    b_ptext_buf[index2] = spu_or(*t3, b_ptext_buf[index2]);
    b_ptext_buf[index2 + 1] = spu_or(*t4, b_ptext_buf[index2 + 1]);
  }
}


// Convert utf8 ciphertext to utf16
static void ConvertCipherText(unsigned int *dst, unsigned char *ciphertext) {
  unsigned int i, temp;
  unsigned char *p_ct = ciphertext;

  for (i = 0; i < 4; ++i) {
    temp  = (atoi16[ARCH_INDEX(p_ct[0])]) << 4;
    temp |= (atoi16[ARCH_INDEX(p_ct[1])]);

    temp |= (atoi16[ARCH_INDEX(p_ct[2])]) << 12;
    temp |= (atoi16[ARCH_INDEX(p_ct[3])]) << 8;

    temp |= (atoi16[ARCH_INDEX(p_ct[4])]) << 20;
    temp |= (atoi16[ARCH_INDEX(p_ct[5])]) << 16;

    temp |= (atoi16[ARCH_INDEX(p_ct[6])]) << 28;
    temp |= (atoi16[ARCH_INDEX(p_ct[7])]) << 24;

    dst[i] = temp;
    p_ct += 8;
  }

  dst[0] -= 0x67452301;
  dst[1] -= 0xefcdab89;
  dst[2] -= 0x98badcfe;
  dst[3] -= 0x10325476;

  // Modify the second element in order to match the expected output
  // of the crypt function.
  dst[1]  = (dst[1] >> 15) | (dst[1] << 17);
  dst[1] -= 0x6ed9eba1 + (dst[2] ^ dst[3] ^ dst[0]);
  dst[1]  = (dst[1] >> 15) | (dst[1] << 17);
  dst[1] -= 0x6ed9eba1;
}


static int ConvertCipherTexts(unsigned int *dst_ct,
			      unsigned char *src_ct[]) {
  unsigned int i;

  for (i = 0; src_ct[i] && i < MAX_CIPHERTEXTS; ++i) {
    ConvertCipherText(&dst_ct[i * 4], src_ct[i]);
  }

  return i;
}


static void DisplayMatchs(unsigned char *results[],
			  unsigned char **ciphertexts) {
  unsigned int i;

  printf("Matched Hashs:\n");
  for (i = 0; i < MAX_CIPHERTEXTS; ++i) {
    if (results[i] != NULL)
      printf("\t%s -> %s\n", ciphertexts[i], results[i]);
  }
}


int main(int argc __attribute__ ((unused)),
	 char **argv __attribute__ ((unused))) {
  unsigned int count = 0, num_ct;
  int num_pt;
  unsigned long long limit = (unsigned long long) -1;
  unsigned short match = 0;
  int (*GeneratePlainText) (unsigned char *, const unsigned int,
			    const unsigned int, void *);
  cmd_opt_t opt;
  void *gen_arg = NULL;

  opt.ciphertext = NULL;
  opt.filename = NULL;
  opt.wordlist = NULL;
  opt = ParseOpt(argc, argv);

  // Must be a multiple of 2
  assert(!(MAX_CIPHERTEXTS % 2));

  // Test mode, only check that tests properly work,
  // ie. check that hashed ciphertext match hash value.
  // Benchmark mode use same ciphertext but generate
  // passwords from charset and use an upper limit.
  if (opt.test || opt.benchmark) {
    unsigned int i;

    // Prepare ciphertexts.
    for (i = 0;
	 tests[i].ciphertext != NULL && i < MAX_CIPHERTEXTS;
	 ++i) {
      unsigned char *ct = (unsigned char *) malloc_align(IN_CT_LENGTH * sizeof(char), 4);
      strncpy((char *) ct, tests[i].ciphertext, IN_CT_LENGTH);
      s_ciphertexts[i] = ct;
    }

    // Number of ciphertexts in test struct must be a multiple of 2.
    assert(!(i % 2));
    s_ciphertexts[i] = NULL;
  }

  // Ciphertexts must be matched immediately.
  if (opt.test) {
    // Assign generator
    GeneratePlainText = &TestGeneratePlainText;
  }

  if (opt.benchmark) {
    // Be sure that those options are not overriden
    opt.start = 0;
    opt.end = -1; // All plaintext space
    opt.charset = 7; // Lower alpha + upper alpha + numeric
    limit = 5000000; // Only limit * NUM_INTER_PT hashes
  }

  // Dummy mode, Generate passwords from charset.
  if (opt.dummy || opt.benchmark) {
    // Initialize generation algorithms and structures
    GenInit(opt.start, opt.end, opt.charset, IN_UTF8_LENGTH);

    // Assign generator
    GeneratePlainText = &DummyGeneratePlainText8;

    if (opt.verbose) {
      // Start
      printf("Start dummy mode from plaintext: ");
      PwdPrint();
    }
  }

  if (opt.wordlist) {
    // Open file
    gen_arg = (void *) WordListOpen(opt.wordlist);

    // Assign generator
    GeneratePlainText = &WordListGeneratePlainText;

    if (opt.verbose)
      // Start
      printf("Start reading %s\n", opt.wordlist);
  }

  // Read ciphertext from command line.
  if ((opt.dummy || opt.wordlist) && opt.ciphertext) {
    unsigned int i;

    assert(MAX_CIPHERTEXTS >= 2);

    // Ciphertext is copied 2 times because we have a constraint
    // to provide a number of ciphertext multiple of 2.
    for (i = 0; i < 2; ++i) {
      unsigned char *ct = (unsigned char *) malloc_align(IN_CT_LENGTH * sizeof(char), 4);
      strncpy((char *) ct, opt.ciphertext, IN_CT_LENGTH);
      s_ciphertexts[i] = ct;
    }

    // Number of ciphertexts in test struct must be a multiple of 2.
    assert(!(i % 2));
    s_ciphertexts[i] = NULL;
  }

  // Read ciphertexts from file.
  if ((opt.dummy || opt.wordlist) && opt.filename) {
    int size, fd;
    unsigned int i;

    fd = open(opt.filename, O_RDONLY);
    if (fd < 0) {
      perror("fopen failed");
      perror(opt.filename);
      exit(1);
    }

    for (i = 0; i < MAX_CIPHERTEXTS; ++i) {
      unsigned char *ct = (unsigned char *) malloc_align(IN_CT_LENGTH * sizeof(char), 4);
      size = read(fd, ct, IN_CT_LENGTH - 1);
      if (size != IN_CT_LENGTH - 1)
	break;
      s_ciphertexts[i] = ct;
      lseek(fd, 1, SEEK_CUR);
    }

    if (i % 2) {
      unsigned char *ct = (unsigned char *) malloc_align(IN_CT_LENGTH * sizeof(char), 4);
      strncpy((char *) ct, (char *) s_ciphertexts[i - 1], IN_CT_LENGTH);
      s_ciphertexts[i++] = ct;
    }

    // Number of ciphertexts in test struct must be a multiple of 2.
    assert(!(i % 2));
    s_ciphertexts[i] = NULL;

    close(fd);
  }


  // Initializations
  jtr_atoi_init();
  num_ct = ConvertCipherTexts(b_ciphertexts, s_ciphertexts);


  // Normal exit conditions:
  //  1- count >= limit: looping counter met a treshold
  //  2- num_pt == 0: no more plaintext to generate from charset space
  //  3- done == 1: all ciphertexts have been matched
  while (++count < limit) {
    // Call to function pointer to generate batch of
    // new plaintexts.
    num_pt = GeneratePlainText(s_ptext_buf, NUM_INTER_PT,
			       IN_UTF8_LENGTH, gen_arg);
    if (num_pt < 0) {
      fprintf(stderr, "GeneratePlainText failure (%d).", num_pt);
      break;
    }

    // Format input plaintexts
    PreparePlainTexts(i_ptext_buf, s_ptext_buf);
    ConvertPlainTexts(i_ptext_buf);

    // Crypt input plaintexts
    NTMD4Crypt();

    // Compare encrypted keys against input ciphertexts
    match = CmpEncryptedShallow(b_ciphertexts, num_ct);
    if (UNLIKELY(match)) {
      // Deep comparison and output formatting
      match = CmpEncryptedDeep(results, s_ptext_buf,
			       b_ciphertexts, num_ct);
      if (LIKELY(match)) {
	unsigned int done = 1, i;

        // Display matched results
	DisplayMatchs(results, s_ciphertexts);

	// See if we're done
	for (i = 0; i < num_ct; ++i)
	  if (results[i] == NULL) {
	    done = 0;
	    break;
	  }

	if (done)
	  // All expected results are matched
	  break;
      }
    }

    if (num_pt != NUM_INTER_PT) {
      if (opt.verbose)
	printf("No more plaintext to generate.\n");
      break;
    }

    if (opt.verbose && !(count % 500000000)) {
      printf("Current plaintext: ");
      PwdPrint();
    }
  }

  // Display vector
  //DisplayOutput();
 
  if (opt.wordlist)
    // Close file
    WordListClose(gen_arg);

  if (opt.verbose && (opt.dummy || opt.benchmark)) {
    printf("End dummy mode to plaintext (excluded): ");
    PwdPrint();
  }

  // Number of plaintexts hashed
  printf("%u plaintexts hashed.\n", count * NUM_INTER_PT);

  // Free data structures
  {
    unsigned int i;

    // Free ciphertexts
    for (i = 0; s_ciphertexts[i] != NULL; ++i)
      free_align(s_ciphertexts[i]);

    // Free results
    for (i = 0; i < MAX_CIPHERTEXTS; ++i)
      if (results[i] != NULL)
	free_align(results[i]);
  }


  return 0;
}
