// Copyright 2008 - SIsécurité
//
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include "parse_opt.h"


static char *TrimWS(char *s) {
  while ((*s != '\0') && (*s == ' '))
    s++;

  return s;
}


static void Usage(char *name) {
  printf("Usage: %s [options]\n", name);
  printf("\t-v verbose mode\n");
  printf("\t-c <ciphertext> ciphertext (string hash) to crack\n");
  printf("\t-f <filename> list of ciphertexts (string hashes "
	 "separated with \\n) to crack\n");
  printf("\t-a <charset> charset corresponding to its #define value\n");
  printf("\t-s <start> generate plaintexts from start position\n");
  printf("\t-e <end> generate plaintexts to end position\n");
  printf("\t-d activate dummy mode\n");
  printf("\t-w <filename> read plaintexts from file (same format "
	 "than password.lst of JtR)\n");
  printf("\t-t activate test mode\n");
  printf("\t-b activate benchmark mode\n");
}


cmd_opt_t ParseOpt(int argc, char **argv) {
  cmd_opt_t opt;
  int optchar = 0;
  unsigned count = 0;

  opt.verbose = 0;
  opt.charset = 0;
  opt.start = 0;
  opt.end = -1;
  opt.dummy = 0;
  opt.test = 0;
  opt.benchmark = 0;
  opt.ciphertext = NULL;
  opt.filename = NULL;
  opt.wordlist = NULL;

  while ((optchar = getopt(argc, argv, "va:c:f:s:e:dtbw:h")) != EOF) {
    switch(optchar) {
    case 'v':
      opt.verbose = 1;
      break;

    case 'a':
      opt.charset = atoi(optarg);
      break;

    case 'c':
      opt.ciphertext = TrimWS(optarg);
      break;

    case 'f':
      opt.filename = TrimWS(optarg);
      break;

    case 's':
      opt.start = atoi(optarg);
      break;

    case 'e':
      opt.end = atoi(optarg);
      break;

    case 'd':
      opt.dummy = 1;
      count++;
      break;

    case 't':
      opt.test = 1;
      count++;
      break;

    case 'b':
      opt.benchmark = 1;
      count++;
      break;

    case 'w':
      opt.wordlist = TrimWS(optarg);
      count++;
      break;

    case 'h':
      Usage(argv[0]);
      exit(0);

    default:
      printf("Unknown option -%c\n\n", optchar);
      Usage(argv[0]);
      exit(1);
    }

  }

  if (count != 1) {
    if (count == 0)
      printf("You must use one option among -b -d -t -w\n\n");
    if (count > 1)
      printf("You must use only one option among -b -d -t -w\n\n");
    Usage(argv[0]);
    exit(1);
  }

  if (opt.dummy && (opt.charset == 0)) {
    printf("Default charset mixed-alpha-num, use -a for "
	   "modifying it\n\n");
    opt.charset = 7;
  }

  if (!opt.test && !opt.benchmark && !opt.dummy &&
      !opt.wordlist) {
    printf("Unspecified mode\n\n");
    Usage(argv[0]);
    exit(1);
  }

  if ((opt.dummy || opt.wordlist) &&
      !opt.ciphertext && !opt.filename) {
    printf("Must provide at least a ciphertext (option -c or -f)\n\n");
    Usage(argv[0]);
    exit(1);
  }

  if (opt.wordlist &&
      (opt.start || (opt.end > -1) || opt.charset))
    printf("Warning: -s or -e or -a options are "
	   "ignored when used with -w\n\n");

  return opt;
}
